/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Define both _XOPEN_SOURCE and _GNU_SOURCE, this makes Solaris happier */
#define _XOPEN_SOURCE 500

/* include lcmaps_c_pep_config.h for PEP_API value, and
 * HAVE_DECL_XACML_RESULT_REMOVEOBLIGATION */
#include "lcmaps_c_pep_config.h"

#if PEP_API == 2 /* Only needed for newer API */
/* Try to provide RTLD_DEFAULT */
/* MacOS needs no POSIX or _DARWIN_C_SOURCE to have RTLD_DEFAULT */
# ifdef __APPLE__
#  define _DARWIN_C_SOURCE
# else
#  define _GNU_SOURCE
# endif
# include <dlfcn.h>
# if !defined(RTLD_DEFAULT) && HAVE_DECL_XACML_RESULT_REMOVEOBLIGATION==1
/* Need to provide xacml_result_removeobligation symbol */
#  include <argus/xacml.h>
# endif
#endif

#include <string.h>
#include <errno.h>

#include <lcmaps/lcmaps_account.h>

#include "pep-c-obligation-handlers.h"
#include "pep-c-obligation-handlers_helpers.h"

#if PEP_API == 1
#define MY_PEP_ERR_OH_PROCESS PEP_ERR_AUTHZ_OH_PROCESS
#else
#define MY_PEP_ERR_OH_PROCESS PEP_ERR_OH_PROCESS
#endif


/* Type of xacml_result_removeobligation() function */
typedef int (*XacmlResultRemoveObligation_t)(xacml_result_t *, int);


/* static pointer to the xacml_result_removeobligation() function, if it exists,
 * or NULL otherwise */
static XacmlResultRemoveObligation_t XacmlResultRemoveObligation = NULL;

/* whether we should ignore unhandled obligations in a Permit result */
static int ignore_unhandled_obligations=0;


/* Sets ignore_unhandled_obligations value */
void set_ignore_unhandled_obligations(int value)    {
    ignore_unhandled_obligations=value;
}

/* returns 0 on success, otherwise -1 */
int check_unhandled_obligations(xacml_response_t * pepc_response)	{
    xacml_result_t * result;
    size_t num_obligations,i;
    const char *obligation_id;
    int rc=0;

    /* When ignore_unhandled_obligations is set, we can return straight away */
    if (ignore_unhandled_obligations)
	return rc;

    /* No support for removing obligations: nothing to check */
    if (XacmlResultRemoveObligation==NULL)  {
	lcmaps_log(LOG_WARNING,
		"%s: PEP-API library does not provide "
		"xacml_result_removeobligation(), cannot check whether all "
		"obligations have been fulfilled.\n", __func__);
	return rc;
    }

    result = xacml_response_getresult(pepc_response,0);
    num_obligations=xacml_result_obligations_length(result);
    if (num_obligations!=0) {
	/* Failed to enforce all obligations */
	rc=-1;
	lcmaps_log (LOG_ERR,
		"%s: Result is Permit but %ld unhandled obligation%s "
		"left in the result:\n",
		__func__, num_obligations,
		(num_obligations==1 ? " is" : "s are"));
	for (i=0; i<num_obligations; i++)   {
	    obligation_id=xacml_obligation_getid(xacml_result_getobligation(result,(int)i));
	    lcmaps_log(LOG_ERR, "%s:    %s\n",
		__func__, obligation_id ? obligation_id : "(NULL)");
	}
    }

    return rc;
}

/* registers the obligation handlers, result is returned in parameters, return
 * value is 0 on success or -1 on error */
int registerObligationHandlers (pep_obligationhandler_t ** oh, size_t * oh_count)
{
    pep_obligationhandler_t *oh_tmp = NULL;
    const size_t oh_max=4;

    /* Initialize oh and oh_count */
    *oh=NULL; *oh_count=0;

    /* Allocate the array for pointers to obligation handlers */
    if ( (oh_tmp = (pep_obligationhandler_t *)calloc (oh_max, sizeof (pep_obligationhandler_t))) == NULL)
	return -1;

    /* Most UNIX now support RTLD_DEFAULT (POSIX reserved) */
#ifdef RTLD_DEFAULT
    /* Clear dlerror() */
    dlerror();
    /* Try to load the function */
    XacmlResultRemoveObligation=
	(XacmlResultRemoveObligation_t)dlsym(RTLD_DEFAULT,
					     "xacml_result_removeobligation");
    if (XacmlResultRemoveObligation==NULL)  {
	/* Do not warn here, we do that at the end in case of a permit */
	lcmaps_log(LOG_DEBUG,
		"%s: PEP-C library does not provide xacml_result_removeobligation() symbol: %s.\n",
		__func__, dlerror());
    }
#elif HAVE_DECL_XACML_RESULT_REMOVEOBLIGATION == 1
    /* No RTLD_DEFAULT but symbol existed at least at build-time */
    lcmaps_log(LOG_DEBUG,
	    "%s: platform does not understand RTLD_DEFAULT, build-time PEP-C library supports xacml_result_removeobligation().\n",
	    __func__);
    XacmlResultRemoveObligation=xacml_result_removeobligation;
#else
    /* No RTLD_DEFAULT and no xacml_result_removeobligation at build-time */
    lcmaps_log(LOG_DEBUG,
	    "%s: plugin is built against PEP-C library without support for xacml_result_removeobligation.\n");
#endif

    /* Setup and link to the UID GID Obligation */
    oh_tmp[0].id      = strdup(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID);
    oh_tmp[0].init    = oh_init_uidgid;
    oh_tmp[0].process = oh_process_uidgid;
    oh_tmp[0].destroy = oh_destroy_uidgid;

    add_supported_obligation_to_list (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID);

    
    /* Setup and link to the Secondary GID Obligation */
    oh_tmp[1].id      = strdup(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs);
    oh_tmp[1].init    = oh_init_secondarygids;
    oh_tmp[1].process = oh_process_secondarygids;
    oh_tmp[1].destroy = oh_destroy_secondarygids;

    add_supported_obligation_to_list (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs);


    /* Setup and link to the Username Obligation */
    oh_tmp[2].id      = strdup(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username);
    oh_tmp[2].init    = oh_init_username;
    oh_tmp[2].process = oh_process_username;
    oh_tmp[2].destroy = oh_destroy_username;

    add_supported_obligation_to_list (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username);

    /* Setup and link to the local-environment-map/posix Obligation */
    oh_tmp[3].id      = strdup(XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX);
    oh_tmp[3].init    = oh_init_local_environment_map_posix;
    oh_tmp[3].process = oh_process_local_environment_map_posix;
    oh_tmp[3].destroy = oh_destroy_local_environment_map_posix;

    add_supported_obligation_to_list (XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX);

    *oh = oh_tmp;
    *oh_count=oh_max;
    
    /* Print the registered supported obligations */
    print_supported_obligations_list();
    lcmaps_log_debug (5, "registerObligationHandlers finished\n");
    
    return 0;
}



int oh_init_uidgid (void)
{
    lcmaps_log_debug (5, "oh_init_uidgid()\n");
    return PEP_OK;
}

int oh_init_secondarygids (void)
{
    lcmaps_log_debug (5, "oh_init_secondarygids()\n");
    return PEP_OK;
}

int oh_init_username (void)
{
    lcmaps_log_debug (5, "oh_init_username()\n");
    return PEP_OK;
}

int oh_init_local_environment_map_posix (void)
{
    lcmaps_log_debug (5, "oh_init_local_environment_map_posix()\n");
    return PEP_OK;
}


int oh_process_uidgid (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, l, results_l, obligations_l, attrs_l, values_l;
    long var_long = -1;
    xacml_response_t * response  = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    const char * obligation_id  = NULL;
    const char * attr_id  = NULL;
    const char *attr_value = NULL;
    int count_oblig_uidgid = 0;
    int count_attrib_uid   = 0;
    int count_attrib_p_gid = 0;
    uid_t uid;
    gid_t gid;

    lcmaps_log_debug (5, "%s\n", __func__);

    if (response == NULL) 
    {
        lcmaps_log_debug (5, "%s: response is NULL\n", __func__);
        return PEP_ERR_NULL_POINTER;
    }


    /* Check response message and verify that the amount of results and the content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response))
    {
	/* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO, "%s: checkResponseSanity() returned a failure condition in the response message. Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    }
    else
    {
        lcmaps_log_debug (5, "%s: checkResponseSanity() returned OK. Continuing with the search for obligation information\n", __func__);
    }

        
    /* Going to obligations blob, skip the rest (as it was tested in detail in checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) 
    {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log_debug (5, "%s: response.result[%u]: %u obligations\n", __func__, i, obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j= 0; j < obligations_l; j++) 
        {
            obligation = xacml_result_getobligation(result,(int)j);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID 
             */
            if (strncmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID, 
                         obligation_id, 
                         strlen(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID)) != 0)
            {
                lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Skipping non-applicable Obligation ID\n",
                                     __func__, i, j, obligation_id);
                continue;
            }


            /* Check if we didn't already had this obligation id... */
            if (count_oblig_uidgid == MAX_APPEARANCE_OF_OBLIGATION_UIDGID)
            {
                lcmaps_log (LOG_ERR, "%s: Error: The result message exceeded the maximum appearance count of the obligation with id: %s.\n", 
                               __func__, XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID);
                return MY_PEP_ERR_OH_PROCESS;
            }
            else
            {
                /* Administrative, counting appearance of obligation-id XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID */
                count_oblig_uidgid++;
            }

            
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Found applicable Obligation ID\n",
                                 __func__, i, j, obligation_id);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].fulfillOn= %s\n",
                                 __func__, i, j, fulfillon_str(xacml_obligation_getfulfillon(obligation))); 

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u]: %d attribute assignments\n",__func__, i, j,(int)attrs_l);

            for (k= 0; k < attrs_l; k++) 
            {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a POSIX Unix UID */
                if (strncmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID, 
                             attr_id, 
                             strlen(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID)
                             ) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    if (count_attrib_uid == MAX_APPEARANCE_OF_ATTRIB_UID)
                    {
                        lcmaps_log (LOG_ERR, "%s: Error: The result message exceeded the maximum appearance count of the attribute id: %s.\n", 
                                       __func__, XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID);
                        return MY_PEP_ERR_OH_PROCESS;
                    }
                    else
                    {
                        /* Administrative, counting appearance of attribute-id XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID */
                        count_attrib_uid++;
                    }


                    /* All ok, adding value */
                    values_l = xacml_attributeassignment_values_length(attr);
                    for (l= 0; l<values_l; l++)
                    {
			attr_value = xacml_attributeassignment_getvalue(attr,l);
                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                             __func__, i, j, k, attr_id);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].value[%d]= %s\n", 
                                             __func__, i, j, k, l, attr_value);
			errno = 0;
                        var_long = strtol (attr_value, NULL, 10);
                        if (errno != 0)
                        {
                            switch (errno)
                            {
                                case EINVAL:
                                case ERANGE:
                                    lcmaps_log (LOG_ERR, "%s: Error: couldn't parse the value at response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n", 
                                                   __func__,  i, j, k, attr_id);
                                    return MY_PEP_ERR_OH_PROCESS;
                            }
                        }
			uid = (uid_t) var_long;
                        lcmaps_log_debug (5, "%s: Adding UID :  %d\n", __func__, uid);
                        if (addCredentialData(UID, &uid)<0) {
			    lcmaps_log (LOG_ERR, "%s: Error: addCredentialData() failed to store uid %u.\n",
				    __func__, uid);
			    return MY_PEP_ERR_OH_PROCESS;
			}
                    }
                }

                /* Is the current value a POSIX Unix GID */
                else if (strncmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID, 
                                  attr_id, 
                                  strlen(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID)
                                  ) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    if (count_attrib_p_gid == MAX_APPEARANCE_OF_ATTRIB_PRI_GID)
                    {
                        lcmaps_log (LOG_ERR, "%s: Error: The result message exceeded the maximum appearance count of the attribute id: %s.\n", 
                                       __func__, XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID);
                        return MY_PEP_ERR_OH_PROCESS;
                    }
                    else
                    {
                        /* Administrative, counting appearance of attribute-id XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID */
                        count_attrib_p_gid++;
                    }

                    /* All ok, adding value */
                    values_l = xacml_attributeassignment_values_length(attr);
                    for (l= 0; l<values_l; l++)
                    {
			attr_value = xacml_attributeassignment_getvalue(attr,l);
                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                             __func__, i, j, k, attr_id);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].value[%d]= %s\n",
                                             __func__, i, j, k, l, attr_value);
			errno = 0;
                        var_long = strtol (attr_value, NULL, 10);
                        if (errno != 0)
                        {
                            switch (errno)
                            {
                                case EINVAL:
                                case ERANGE:
                                    lcmaps_log (LOG_ERR, "%s: Error: couldn't parse the value at response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n", 
                                                   __func__,  i, j, k, attr_id);
                                    return MY_PEP_ERR_OH_PROCESS;
                            }
                        }
			gid = (gid_t) var_long;
                        lcmaps_log_debug (5, "%s: Adding GID :  %d\n", __func__, gid);
                        if (addCredentialData(PRI_GID, &gid)<0) {
			    lcmaps_log (LOG_ERR, "%s: Error: addCredentialData() failed to store gid %u.\n",
				    __func__, gid);
			    return MY_PEP_ERR_OH_PROCESS;
			}
                    }
                }
                else
                {
                    lcmaps_log_debug (1, "%s: Error: Unknown attribute found in the obligation block: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                   __func__, i, j, k, attr_id);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation and update number of
	     * obligations */
	    if (XacmlResultRemoveObligation)	{
		XacmlResultRemoveObligation(result,(int)j);
		obligations_l--;
	    }
        }
    }

    return PEP_OK;
}

int oh_process_secondarygids (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, l, results_l, obligations_l, attrs_l, values_l;
    xacml_response_t * response  = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    const char * obligation_id  = NULL;
    const char * attr_id  = NULL;
    const char *attr_value = NULL;
    int count_attrib_gid = 0;
    long var_long;
    gid_t gid;

    lcmaps_log_debug (5, "%s\n", __func__);

    if (response == NULL) 
    {
        lcmaps_log_debug (5, "%s: response is NULL\n", __func__);
        return 1;
    }


    /* Check response message and verify that the amount of results and the content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response))
    {
        /* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO, "%s: checkResponseSanity() returned a failure condition in the response message. Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    }
    else
    {
        lcmaps_log_debug (5, "%s: checkResponseSanity() returned OK. Continuing with the search for obligation information\n", __func__);
    }

        
    /* Going to obligations blob, skip the rest (as it was tested in detail in checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) 
    {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log_debug (5, "%s: response.result[%u]: %d obligations\n", __func__, i, (int)obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j= 0; j < obligations_l; j++) 
        {
            obligation = xacml_result_getobligation(result,(int)j);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs
             */
            if (strncmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs,
                         obligation_id, 
                         strlen(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs)) != 0)
            {
                lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Skipping non-applicable Obligation ID\n",
                                     __func__, i, j, obligation_id);
                continue;
            }

            /* We've gottum! */
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Found applicable Obligation ID\n",
                                 __func__, i, j, obligation_id);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].fulfillOn= %s\n",
                                 __func__, i, j, fulfillon_str(xacml_obligation_getfulfillon(obligation))); 

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u]: %d attribute assignments\n",__func__, i, j,(int)attrs_l);

            for (k= 0; k < attrs_l; k++) 
            {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a POSIX Unix GID */
                if (strncmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID, 
                             attr_id, 
                             strlen(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID)
                             ) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    if (count_attrib_gid == MAX_APPEARANCE_OF_ATTRIB_SEC_GID)
                    {
                        lcmaps_log (LOG_ERR, "%s: Error: The result message exceeded the maximum appearance count of the attribute id: %s.\n", 
                                       __func__, XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID);
                        return MY_PEP_ERR_OH_PROCESS;
                    }
                    else
                    {
                        /* Administrative, counting appearance of attribute-id XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID */
                        count_attrib_gid++;
                    }

                    values_l = xacml_attributeassignment_values_length(attr);
                    for (l= 0; l<values_l; l++)
                    {
			attr_value = xacml_attributeassignment_getvalue(attr,l);
                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                             __func__, i, j, k, attr_id);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].value[%d]= %s\n",
                                             __func__, i, j, k, l, attr_value);
			errno = 0;
                        var_long = strtol (attr_value, NULL, 10);
                        if (errno != 0)
                        {
                            switch (errno)
                            {
                                case EINVAL:
                                case ERANGE:
                                    lcmaps_log (LOG_ERR, "%s: Error: couldn't parse the value at response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n", 
                                                   __func__,  i, j, k, attr_id);
                                    return MY_PEP_ERR_OH_PROCESS;
                            }
                        }
			gid = (gid_t) var_long;
                        lcmaps_log_debug (5, "%s: Adding Sec GID :  %d\n", __func__, gid);
                        if (addCredentialData(SEC_GID, &gid)<0)	{
			    lcmaps_log (LOG_ERR, "%s: Error: addCredentialData() failed to store sgid %u.\n",
				    __func__, gid);
			    return MY_PEP_ERR_OH_PROCESS;
			}
                    }
                }
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: in processing the unexpected XACML-attribute: \"%s\"\n", __func__, XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation and update number of
	     * obligations */
	    if (XacmlResultRemoveObligation)	{
		XacmlResultRemoveObligation(result,(int)j);
		obligations_l--;
	    }
        }
    }

    return PEP_OK;
}

int oh_process_username (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, l, results_l, obligations_l, attrs_l, values_l;
    xacml_response_t * response  = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    const char * obligation_id  = NULL;
    const char * attr_id  = NULL;
    int count_attrib_username  = 0;
    const char * username = NULL;

    lcmaps_log_debug (5, "%s\n", __func__);

    if (response == NULL) 
    {
        lcmaps_log_debug (5, "%s: response is NULL\n", __func__);
        return 1;
    }


    /* Check response message and verify that the amount of results and the content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response))
    {
        /* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO, "%s: checkResponseSanity() returned a failure condition in the response message. Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    }
    else
    {
        lcmaps_log_debug (5, "%s: checkResponseSanity() returned OK. Continuing with the search for obligation information\n", __func__);
    }

        
    /* Going to obligations blob, skip the rest (as it was tested in detail in checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) 
    {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log_debug (5, "%s: response.result[%u]: %d obligations\n", __func__, i, (int)obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j= 0; j < obligations_l; j++) 
        {
            obligation = xacml_result_getobligation(result,(int)j);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username     
             */
            if (strncmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username,
                         obligation_id, 
                         strlen(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username)) != 0)
            {
                lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Skipping non-applicable Obligation ID\n",
                                     __func__, i, j, obligation_id);
                continue;
            }

            /* We've gottum! */
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Found applicable Obligation ID\n",
                                 __func__, i, j, obligation_id);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].fulfillOn= %s\n",
                                 __func__, i, j, fulfillon_str(xacml_obligation_getfulfillon(obligation))); 

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u]: %d attribute assignments\n",__func__, i, j,(int)attrs_l);

            for (k= 0; k < attrs_l; k++) 
            {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a Username */
                if (strncmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME, 
                             attr_id, 
                             strlen(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME)
                             ) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    /* The maximum number of Usernames in an obligation is 1, and only 1 */
                    if (count_attrib_username == 1)
                    {
                        lcmaps_log (LOG_ERR, "%s: Error: The result message exceeded the maximum appearance count of the attribute id: %s.\n", 
                                       __func__, XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME);
                        return MY_PEP_ERR_OH_PROCESS;
                    }
                    else
                    {
                        /* Administrative, counting appearance of attribute-id XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID */
                        count_attrib_username++;
                    }

                    values_l = xacml_attributeassignment_values_length(attr);
                    for (l= 0; l<values_l; l++)
                    {
                        /* Got the username, now to convert it into UIDs and GIDs through the POSIX account system */
                        username = xacml_attributeassignment_getvalue(attr,l);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                             __func__, i, j, k, attr_id);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].value[%d]= %s\n",
                                             __func__, i, j, k, l, username);

                        /* Converting the Username into (local) numerical UID, GID and Secondary GIDs
                         * and registering the values into the LCMAPS framework */
                        if (addCredentialDataFromUsername (username) != 0)
                        {
                            lcmaps_log (LOG_ERR, "%s: Error: in adding Username information.\n", __func__);
                            return MY_PEP_ERR_OH_PROCESS;
                        }
                    }
                }
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: Unexpected attribute found with the ID: %s\n", __func__, attr_id);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation and update number of
	     * obligations */
	    if (XacmlResultRemoveObligation)	{
		XacmlResultRemoveObligation(result,(int)j);
		obligations_l--;
	    }
        }
    }

    return PEP_OK;
}

int oh_process_local_environment_map_posix (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, l, results_l, obligations_l, attrs_l, values_l;
    xacml_response_t * response = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    const char * obligation_id  = NULL;
    const char * attr_id        = NULL;
    int count_attrib_userid     = 0;
    int count_attrib_groupid    = 0;
    int count_attrib_secgroupid = 0;
    const char * userid         = NULL;
    const char * groupid        = NULL;
    const char * secgroupid     = NULL;
    uid_t this_uid;
    gid_t this_prigid;
    gid_t this_secgid;
    /* gid_t * gid                 = NULL; */
    /* int     cnt_gid             = 0; */

    lcmaps_log_debug (5, "%s\n", __func__);

    if (response == NULL) 
    {
        lcmaps_log_debug (5, "%s: response is NULL\n", __func__);
        return 1;
    }


    /* Check response message and verify that the amount of results and the content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response))
    {
	/* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO, "%s: checkResponseSanity() returned a failure condition in the response message. Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    }
    else
    {
        lcmaps_log_debug (5, "%s: checkResponseSanity() returned OK. Continuing with the search for obligation information\n", __func__);
    }

        
    /* Going to obligations blob, skip the rest (as it was tested in detail in checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) 
    {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log_debug (5, "%s: response.result[%u]: %d obligations\n", __func__, i, (int)obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j= 0; j < obligations_l; j++) 
        {
            obligation = xacml_result_getobligation(result,(int)j);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX 
             */
            if (strncmp (XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX,
                         obligation_id, 
                         strlen(XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX)) != 0)
            {
                lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Skipping non-applicable Obligation ID\n",
                                     __func__, i, j, obligation_id);
                continue;
            }

            /* We've gottum! */
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].id = \"%s\". Found applicable Obligation ID\n",
                                 __func__, i, j, obligation_id);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].fulfillOn= %s\n",
                                 __func__, i, j, fulfillon_str(xacml_obligation_getfulfillon(obligation))); 

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u]: %d attribute assignments\n",__func__, i, j,(int)attrs_l);

            for (k= 0; k < attrs_l; k++) 
            {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a User-ID */
                if (strncmp (XACML_GRIDWN_ATTRIBUTE_USER_ID, 
                             attr_id, 
                             strlen(XACML_GRIDWN_ATTRIBUTE_USER_ID)
                             ) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    /* The maximum number of Usernames in an obligation is 1, and only 1 */
                    if (count_attrib_userid == 1)
                    {
                        lcmaps_log (LOG_ERR, "%s: Error: The result message exceeded the maximum appearance count of the attribute id: %s.\n", 
                                       __func__, XACML_GRIDWN_ATTRIBUTE_USER_ID);
                        return MY_PEP_ERR_OH_PROCESS;
                    }
                    else
                    {
                        /* Administrative, counting appearance of attribute-id XACML_GRIDWN_ATTRIBUTE_USER_ID */
                        count_attrib_userid++;
                    }

                    values_l = xacml_attributeassignment_values_length(attr);
                    for (l= 0; l<values_l; l++)
                    {
                        /* Got the username, now to convert it into UIDs through the POSIX account system */
                        userid = xacml_attributeassignment_getvalue(attr,l);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                             __func__, i, j, k, attr_id);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].value[%d]= %s\n",
                                             __func__, i, j, k, l, userid);

                        /* Converting the User-ID into a UID
                         * and registering the values into the LCMAPS framework */
                        if (threadsafe_getuid_from_name (userid, &this_uid) < 0)
                        {
                            lcmaps_log (LOG_ERR, "%s: Error: couldn't find the account entry for the User-ID \"%s\".\n", __func__, userid);
                            return MY_PEP_ERR_OH_PROCESS;
                        }
			lcmaps_log_debug (5, "%s: Register UID %d in LCMAPS\n", __func__, this_uid);
			if (addCredentialData(UID, &this_uid)<0)	{
			    lcmaps_log (LOG_ERR, "%s: Error: addCredentialData() failed to store uid %u.\n",
				    __func__, this_uid);
			    return MY_PEP_ERR_OH_PROCESS;
			}
                    }
                }
                /* Is the current value a Primary Group-ID */
                else if (strncmp (XACML_GRIDWN_ATTRIBUTE_GROUP_ID_PRIMARY, 
                                  attr_id, 
                                  strlen(XACML_GRIDWN_ATTRIBUTE_GROUP_ID_PRIMARY)
                                 ) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    /* The maximum number of Primary Groups in an obligation is 1, and only 1 */
                    if (count_attrib_groupid == 1)
                    {
                        lcmaps_log (LOG_ERR, "%s: Error: The result message exceeded the maximum appearance count of the attribute id: %s.\n", 
                                       __func__, XACML_GRIDWN_ATTRIBUTE_GROUP_ID_PRIMARY);
                        return MY_PEP_ERR_OH_PROCESS;
                    }
                    else
                    {
                        /* Administrative, counting appearance of attribute-id XACML_GRIDWN_ATTRIBUTE_GROUP_ID */
                        count_attrib_groupid++;
                    }

                    values_l = xacml_attributeassignment_values_length(attr);
                    for (l= 0; l<values_l; l++)
                    {
                        /* Got the groupname, now to convert it into a GID through the POSIX account system */
                        groupid = xacml_attributeassignment_getvalue(attr,l);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                             __func__, i, j, k, attr_id);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].value[%d]= %s\n",
                                             __func__, i, j, k, l, groupid);

                        /* Converting the Group-ID into a GID
                         * and registering the values into the LCMAPS framework */
                        if (threadsafe_getgid_from_name (groupid, &this_prigid)<0)
                        {
                            lcmaps_log (LOG_ERR, "%s: Error: couldn't find the group entry for the Group-ID \"%s\".\n", __func__, groupid);
                            return MY_PEP_ERR_OH_PROCESS;
                        }
			lcmaps_log_debug (5, "%s: Register Primary GID %d in LCMAPS\n", __func__, this_prigid);
                        if (addCredentialData(PRI_GID, &this_prigid)<0)	{
			    lcmaps_log (LOG_ERR, "%s: Error: addCredentialData() failed to store pgid %u.\n",
				    __func__, this_prigid);
			    return MY_PEP_ERR_OH_PROCESS;
			}
                    }
                }
                /* Is the current value a Secondary Group-ID */
                else if (strncmp (XACML_GRIDWN_ATTRIBUTE_GROUP_ID,
                                  attr_id, 
                                  strlen(XACML_GRIDWN_ATTRIBUTE_GROUP_ID)
                                 ) == 0)
                {
                    /* Administrative, counting appearance of attribute-id XACML_GRIDWN_ATTRIBUTE_GROUP_ID_PRIMARY */
                    count_attrib_secgroupid++;

                    values_l = xacml_attributeassignment_values_length(attr);
                    for (l= 0; l<values_l; l++)
                    {
                        /* Got the groupname, now to convert it into a GID through the POSIX account system */
                        secgroupid = xacml_attributeassignment_getvalue(attr,l);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].id= %s\n",
                                             __func__, i, j, k, attr_id);

                        lcmaps_log_debug (5, "%s: response.result[%u].obligation[%u].attributeassignment[%d].value[%d]= %s\n",
                                             __func__, i, j, k, l, secgroupid);

                        /* Converting the Group-ID into a GID
                         * and registering the values into the LCMAPS framework */
                        if ( threadsafe_getgid_from_name (secgroupid, &this_secgid)<0)
                        {
                            lcmaps_log (LOG_ERR, "%s: Error: couldn't find the group entry for the Group-ID \"%s\".\n", __func__, secgroupid);
                            return MY_PEP_ERR_OH_PROCESS;
                        }
			lcmaps_log_debug (5, "%s: Register Secondary GID %d in LCMAPS\n", __func__, this_secgid);
			if (addCredentialData(SEC_GID, &this_secgid)<0)	{
			    lcmaps_log (LOG_ERR, "%s: Error: addCredentialData() failed to store sgid %u.\n",
				    __func__, this_secgid);
			    return MY_PEP_ERR_OH_PROCESS;
			}
                    }
                }
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: Unexpected attribute found with the ID: %s\n", __func__, attr_id);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation and update number of
	     * obligations */
	    if (XacmlResultRemoveObligation)	{
		XacmlResultRemoveObligation(result,(int)j);
		obligations_l--;
	    }
        }
    }

    return PEP_OK;
}

int oh_destroy_uidgid (void)
{
    lcmaps_log_debug (5, "oh_destroy_uidgid()\n");
    return PEP_OK;
}

int oh_destroy_secondarygids (void)
{
    lcmaps_log_debug (5, "oh_destroy_secondarygids()\n");
    return PEP_OK;
}

int oh_destroy_username(void)
{
    lcmaps_log_debug (5, "oh_destroy_username()\n");
    return PEP_OK;
}

int oh_destroy_local_environment_map_posix (void)
{
    lcmaps_log_debug (5, "oh_destroy_local_environment_map_posix()\n");
    return PEP_OK;
}




